=== modified file 'Makefile'
--- Makefile	2012-01-10 11:16:56 +0000
+++ Makefile	2012-01-11 08:06:03 +0000
@@ -1,2 +1,4 @@
 check:
 	python -m unittest grackle.tests.test_client
+
+.PHONY: check

=== modified file 'grackle-put-message'
--- grackle-put-message	2012-01-10 11:06:36 +0000
+++ grackle-put-message	2012-01-13 13:52:09 +0000
@@ -6,9 +6,10 @@
     """An error caused by a user mistake."""
 
 def main(*args):
-    if len(args) != 1:
+    if len(args) != 2:
         raise UserError("Wrong number of arguments supplied.")
-    client.put_message(args[0], sys.stdin)
+    grackle = client.GrackleClient('localhost', 8123)
+    grackle.put_message(args[0], args[1], sys.stdin)
 
 
 if __name__ == '__main__':

=== modified file 'grackle/client.py'
--- grackle/client.py	2012-01-10 14:41:33 +0000
+++ grackle/client.py	2012-01-13 15:39:22 +0000
@@ -1,10 +1,14 @@
 import httplib
+import simplejson
 from urlparse import urlunparse
-from urllib import quote
-
-
-def message_path(archive_name, message_id):
-    return '/%s/%s' % (quote(archive_name), quote(message_id))
+from urllib import (
+    quote,
+    urlencode,
+)
+
+
+class UnsupportedOrder(Exception):
+    pass
 
 
 class GrackleClient:
@@ -12,17 +16,25 @@
     def __init__(self, host, port):
         self.host = host
         self.port = port
-
-    def message_url(self, archive_name, message_id):
-        netloc = '%s:%d' % (self.host, self.port)
-        path = message_path(archive_name, message_id)
-        return urlunparse(('http', netloc, path, '', '', ''))
-
-    def put_message(self, archive_name, message_id, file_obj):
-        url = self.message_url(archive_name, message_id)
-        connection = httplib.HTTPConnection(self.host, self.port)
-        connection.request('PUT', url, file_obj.read())
-        response = connection.getresponse()
+        self.netloc = '%s:%d' % (host, port)
+
+    def archive_url(self, archive_id, query):
+        path = '/archive/%s' % quote(archive_id)
+        query_string = urlencode(query)
+        return urlunparse(('http', self.netloc, path, '', query_string, ''))
+
+    def _get_connection(self):
+        return httplib.HTTPConnection(self.host, self.port)
+
+    def _verb_archive(self, verb, archive_id, query, body=None):
+        url = self.archive_url(archive_id, query)
+        connection = self._get_connection()
+        connection.request(verb, url, body)
+        return connection.getresponse()
+
+    def put_message(self, archive_id, key, file_obj):
+        response = self._verb_archive(
+            'POST', archive_id, {'key': key}, file_obj.read())
         data = response.read()
         if response.status == httplib.BAD_REQUEST:
             raise Exception('wtf')
@@ -30,3 +42,28 @@
             return
         else:
             raise Exception('!!')
+
+    def get_messages(self, archive_id, message_ids=None, limit=None,
+                     memo=None, order=None, headers=None,
+                     max_body_length=None, include_hidden=False):
+        parameters = {}
+        if message_ids is not None:
+            parameters['message_ids'] = message_ids
+        if limit is not None:
+            parameters['limit'] = limit
+        if memo is not None:
+            parameters['memo'] = memo
+        if order is not None:
+            parameters['order'] = order
+        if headers is not None:
+            parameters['headers'] = headers
+        if max_body_length is not None:
+            parameters['max_body_length'] = max_body_length
+        parameters['include_hidden'] = include_hidden
+        query = {'parameters': simplejson.dumps(parameters)}
+        response = self._verb_archive('GET', archive_id, query)
+        if response.status == httplib.BAD_REQUEST:
+            raise UnsupportedOrder
+        data = response.read()
+        return simplejson.loads(data)
+

=== modified file 'grackle/tests/test_client.py'
--- grackle/tests/test_client.py	2012-01-10 14:44:37 +0000
+++ grackle/tests/test_client.py	2012-01-13 15:41:11 +0000
@@ -5,39 +5,150 @@
 import httplib
 import os
 from signal import SIGKILL
+import simplejson
 from StringIO import StringIO
 from unittest import TestCase
+from urlparse import urlparse
+from urlparse import parse_qs
 
 from testtools import ExpectedException
 
 from grackle.client import (
     GrackleClient,
-    message_path,
+    UnsupportedOrder,
     )
 
 
-class Forked:
-
-    def __init__(self, func_or_method, *args):
-        self.func_or_method = func_or_method
+def threaded_messages(messages):
+    threads = {}
+    count = 0
+    pending = []
+    for message in messages:
+        if message.get('in_reply_to') is None:
+            threads[message['message_id']] = [message]
+            count += 1
+        else:
+            pending.append(message)
+    for message in pending:
+        threads[message['in_reply_to']].append(message)
+    return threads.values()
+
+
+class GrackleStore:
+
+    def __init__(self, messages):
+        self.messages = messages
+
+    def get_messages(self, archive_id, query_string):
+        query = parse_qs(query_string)
+        parameters = simplejson.loads(query['parameters'][0])
+        order = parameters.get('order')
+        messages = self.messages[archive_id]
+        if order is not None :
+            if order not in SUPPORTED_ORDERS:
+                raise UnsupportedOrder
+            elif order.startswith('thread_'):
+                threaded = threaded_messages(messages)
+                messages = []
+                if order == 'thread_subject':
+                    threaded.sort(key=lambda t: t[0]['subject'])
+                if order == 'thread_oldest':
+                    threaded.sort(key=lambda t: min(m['date'] for m in t))
+                if order == 'thread_newest':
+                    threaded.sort(key=lambda t: max(m['date'] for m in t))
+                for thread in threaded:
+                    messages.extend(thread)
+            else:
+                messages.sort(key=lambda m: m[order])
+        new_messages = []
+        for message in messages:
+            if (
+                not parameters['include_hidden']
+                and message.get('hidden', False)):
+                continue
+
+            if ('message_ids' in parameters and
+                message['message_id'] not in parameters['message_ids']):
+                continue
+            message = dict(message)
+            if 'headers' in parameters:
+                headers = dict(
+                    (k, v) for k, v in message['headers'].iteritems()
+                    if k in parameters['headers'])
+                message['headers'] = headers
+            max_body = parameters.get('max_body_length')
+            if max_body is not None:
+                message['body'] = message['body'][:max_body]
+            new_messages.append(message)
+        messages = new_messages
+        limit = parameters.get('limit', 100)
+        memo = parameters.get('memo')
+        message_id_indices = dict(
+            (m['message_id'], idx) for idx, m in enumerate(messages))
+        if memo is None:
+            start = 0
+        else:
+            start = message_id_indices[memo.encode('rot13')]
+        if start > 0:
+            previous_memo = messages[start - 1]['message_id'].encode('rot13')
+        else:
+            previous_memo = None
+        end = min(start + limit, len(messages))
+        if end < len(messages):
+            next_memo = messages[end]['message_id'].encode('rot13')
+        else:
+            next_memo = None
+        messages = messages[start:end]
+
+        response = {
+            'messages': messages,
+            'next_memo': next_memo,
+            'previous_memo': previous_memo
+            }
+        return response
+
+
+
+class ForkedFake:
+
+    def __init__(self, port, messages=None):
         self.pid = None
-        self.args = args
+        self.port = port
+        self.messages = messages
+        self.read_end, self.write_end = os.pipe()
+
+    def is_ready(self):
+        os.write(self.write_end, 'asdf')
 
     def __enter__(self):
         pid = os.fork()
-        if pid != 0:
-            self.pid = pid
-            return
-        self.func_or_method(*self.args)
+        if pid == 0:
+            self.start_server()
+        self.pid = pid
+        os.read(self.read_end, 1)
+        return
 
+    def start_server(self):
+        service = HTTPServer(('', self.port), FakeGrackleRequestHandler)
+        service.store = GrackleStore(self.messages)
+        for archive_id, messages in service.store.messages.iteritems():
+            for message in messages:
+                message.setdefault('headers', {})
+        self.is_ready()
+        service.serve_forever()
 
     def __exit__(self, exc_type, exc_val, traceback):
         os.kill(self.pid, SIGKILL)
 
 
+SUPPORTED_ORDERS = set(
+    ['date', 'author', 'subject', 'thread_newest', 'thread_oldest',
+     'thread_subject'])
+
+
 class FakeGrackleRequestHandler(BaseHTTPRequestHandler):
 
-    def do_PUT(self):
+    def do_POST(self):
         message = self.rfile.read(int(self.headers['content-length']))
         if message == 'This is a message':
             self.send_response(httplib.CREATED)
@@ -46,25 +157,211 @@
         else:
             self.send_error(httplib.BAD_REQUEST)
 
-
-def run_service(port):
-    service = HTTPServer(('', port), FakeGrackleRequestHandler)
-    service.serve_forever()
-
+    def do_GET(self):
+        scheme, netloc, path, params, query_string, fragments = (
+            urlparse(self.path))
+        parts = path.split('/')
+        if parts[1] == 'archive':
+            try:
+                response = self.server.store.get_messages(
+                    parts[2], query_string)
+                self.send_response(httplib.OK)
+                self.end_headers()
+                self.wfile.write(simplejson.dumps(response))
+            except UnsupportedOrder:
+                self.send_response(httplib.BAD_REQUEST)
+                self.wfile.write('Unsupported order')
+                return
+
+
+def fake_grackle_service(client, messages=None):
+    if messages is None:
+        messages = {}
+    return ForkedFake(client.port, messages)
 
 
 class TestPutMessage(TestCase):
 
     def test_put_message(self):
-        client = GrackleClient('localhost', 8435)
-        with Forked(run_service, client.port):
+        client = GrackleClient('localhost', 8436)
+        with fake_grackle_service(client):
             client.put_message('arch1', 'asdf', StringIO('This is a message'))
             with ExpectedException(Exception, 'wtf'):
                 client.put_message('arch1', 'asdf',
                     StringIO('This is not a message'))
 
 
-class TestMessagePath(TestCase):
-
-    def test_message_path(self):
-        self.assertEqual('/my%20/%3C', message_path('my ', '<'))
+class TestGetMessages(TestCase):
+
+    def assertIDOrder(self, ids, messages):
+        self.assertEqual(ids, [m['message_id'] for m in messages])
+
+    def assertMessageIDs(self, ids, messages):
+        self.assertIDOrder(
+            sorted(ids), sorted(messages, key=lambda m:m['message_id']))
+
+    def test_get_messages(self):
+        client = GrackleClient('localhost', 8435)
+        with fake_grackle_service(client,
+            {'baz':
+            [{'message_id': 'foo'},
+             {'message_id': 'bar'}]}):
+            response = client.get_messages('baz')
+        self.assertEqual(['bar', 'foo'], sorted(m['message_id'] for m in
+            response['messages']))
+        self.assertIs(None, response['next_memo'])
+        self.assertIs(None, response['previous_memo'])
+
+    def test_get_messages_by_id(self):
+        client = GrackleClient('localhost', 8437)
+        with fake_grackle_service(client,
+            {'baz':
+            [{'message_id': 'foo'},
+             {'message_id': 'bar'}]}):
+            response = client.get_messages('baz', message_ids=['foo'])
+        message, = response['messages']
+        self.assertEqual('foo', message['message_id'])
+
+    def test_get_messages_batching(self):
+        client = GrackleClient('localhost', 8438)
+        with fake_grackle_service(client,
+            {'baz':
+            [{'message_id': 'foo'},
+             {'message_id': 'bar'}]}):
+            response = client.get_messages('baz', limit=1)
+            self.assertEqual(1, len(response['messages']))
+            messages = response['messages']
+            response = client.get_messages(
+                'baz', limit=1, memo=response['next_memo'])
+            self.assertEqual(1, len(response['messages']))
+            messages.extend(response['messages'])
+            self.assertMessageIDs(['foo', 'bar'], messages)
+
+    def get_messages_member_order_test(self, key):
+        client = GrackleClient('localhost', 8439)
+        with fake_grackle_service(client,
+                {'baz': [{'message_id': 'foo', key: '2011-03-25'},
+                 {'message_id': 'bar', key: '2011-03-24'}]}):
+            response = client.get_messages('baz')
+            self.assertIDOrder(['foo', 'bar'], response['messages'])
+            response = client.get_messages('baz', order=key)
+            self.assertIDOrder(['bar', 'foo'], response['messages'])
+
+    def test_get_messages_date_order(self):
+        self.get_messages_member_order_test('date')
+
+    def test_get_messages_author_order(self):
+        self.get_messages_member_order_test('author')
+
+    def test_get_messages_subject_order(self):
+        self.get_messages_member_order_test('subject')
+
+    def test_get_messages_thread_subject_order(self):
+        client = GrackleClient('localhost', 8439)
+        with fake_grackle_service(client, {'baz': [
+            {'message_id': 'bar', 'subject': 'y'},
+            {'message_id': 'qux', 'subject': 'z'},
+            {'message_id': 'foo', 'subject': 'x', 'in_reply_to': 'qux'},
+            ]}):
+            response = client.get_messages('baz')
+            self.assertIDOrder(['bar', 'qux', 'foo'], response['messages'])
+            response = client.get_messages('baz', order='subject')
+            self.assertIDOrder(['foo', 'bar', 'qux'], response['messages'])
+            response = client.get_messages('baz', order='thread_subject')
+            self.assertIDOrder(['bar', 'qux', 'foo'], response['messages'])
+
+    def test_get_messages_thread_oldest_order(self):
+        client = GrackleClient('localhost', 8439)
+        with fake_grackle_service(client, {'baz': [
+            {'message_id': 'bar', 'date': 'x'},
+            {'message_id': 'qux', 'date': 'z'},
+            {'message_id': 'foo', 'date': 'y', 'in_reply_to': 'qux'},
+            ]}):
+            response = client.get_messages('baz')
+            self.assertIDOrder(['bar', 'qux', 'foo'], response['messages'])
+            response = client.get_messages('baz', order='date')
+            self.assertIDOrder(['bar', 'foo', 'qux'], response['messages'])
+            response = client.get_messages('baz', order='thread_oldest')
+            self.assertIDOrder(['bar', 'qux', 'foo'], response['messages'])
+
+    def test_get_messages_thread_newest_order(self):
+        client = GrackleClient('localhost', 8439)
+        with fake_grackle_service(client, {'baz': [
+            {'message_id': 'bar', 'date': 'x'},
+            {'message_id': 'qux', 'date': 'w'},
+            {'message_id': 'foo', 'date': 'y', 'in_reply_to': 'bar'},
+            {'message_id': 'baz', 'date': 'z', 'in_reply_to': 'qux'},
+            ]}):
+            response = client.get_messages('baz', order='date')
+            self.assertIDOrder(
+                ['qux', 'bar', 'foo', 'baz'], response['messages'])
+            response = client.get_messages('baz', order='thread_newest')
+            self.assertIDOrder(
+                ['bar', 'foo', 'qux', 'baz'], response['messages'])
+
+    def test_get_messages_unsupported_order(self):
+        client = GrackleClient('localhost', 8439)
+        with fake_grackle_service(client,
+                {'baz': [{'message_id': 'foo', 'date': '2011-03-25'},
+                 {'message_id': 'bar', 'date': '2011-03-24'}]}):
+            with ExpectedException(UnsupportedOrder):
+                client.get_messages('baz', order='nonsense')
+
+    def test_get_messages_headers_no_headers(self):
+        client = GrackleClient('localhost', 8440)
+        with fake_grackle_service(client,
+            {'baz': [
+                {'message_id': 'foo'}
+            ]}):
+            response = client.get_messages('baz', headers=[
+                'Subject', 'Date', 'X-Launchpad-Message-Rationale'])
+        first_message = response['messages'][0]
+        self.assertEqual('foo', first_message['message_id'])
+        self.assertEqual({}, first_message['headers'])
+
+    def test_get_messages_headers_exclude_headers(self):
+        client = GrackleClient('localhost', 8441)
+        with fake_grackle_service(client,
+            {'baz': [
+                {'message_id': 'foo', 'headers': {'From': 'me'}}
+            ]}):
+            response = client.get_messages('baz', headers=[
+                'Subject', 'Date', 'X-Launchpad-Message-Rationale'])
+        first_message = response['messages'][0]
+        self.assertEqual('foo', first_message['message_id'])
+        self.assertEqual({}, first_message['headers'])
+
+    def test_get_messages_headers_include_headers(self):
+        client = GrackleClient('localhost', 8442)
+        with fake_grackle_service(client,
+            {'baz': [
+                {'message_id': 'foo', 'headers': {'From': 'me', 'To': 'you'}}
+            ]}):
+            response = client.get_messages('baz', headers=[
+                'From', 'To'])
+        first_message = response['messages'][0]
+        self.assertEqual('foo', first_message['message_id'])
+        self.assertEqual({'From': 'me', 'To': 'you'}, first_message['headers'])
+
+    def test_get_messages_max_body_length(self):
+        client = GrackleClient('localhost', 8443)
+        with fake_grackle_service(client,
+            {'baz': [
+                {'message_id': 'foo', 'body': u'abcdefghi'}
+            ]}):
+            response = client.get_messages('baz', max_body_length=3)
+        first_message = response['messages'][0]
+        self.assertEqual('abc', first_message['body'])
+
+    def test_include_hidden(self):
+        client = GrackleClient('localhost', 8444)
+        with fake_grackle_service(client,
+            {'baz': [
+                {'message_id': 'foo', 'hidden': True},
+                {'message_id': 'bar', 'hidden': False}
+            ]}):
+            response = client.get_messages('baz', include_hidden=True)
+            self.assertMessageIDs(['bar', 'foo'], response['messages'])
+            response = client.get_messages('baz', include_hidden=False)
+            self.assertMessageIDs(['bar'], response['messages'])
+

